# -*- coding: utf-8 -*-
"""
Created on Mon May 01 13:14:08 2017

@author: Oscar
"""
from os import path
import os
from exceptions import AppException

class FrocCurve() :
  def __init__(self) :
    self.TPsFPsList = []
    self.numFalseNegatives = None
    self._frocCurve = []
    self._rfrocCurve = []
    self.frocFileName = path.join(os.getcwd(), 'froc.csv')

  def addPoint(self, likelihood, truthUID) :
    p = {}
    p['likelihood'] = likelihood
    p['truthUID'] = truthUID # None means it is a false positive.
    self.TPsFPsList.append(p)

  def getAllTruthUIDs(self) :
    truthUIDs = set()
    ### Add true positives:
    for p in self.TPsFPsList :
      if p['truthUID'] != None :
        truthUIDs.add(p['truthUID'])
    ### Add false negatives:
    ### TO DO
    return list(truthUIDs)

  def getNumTPsIncludingDuplicates(self) :
    n = 0
    for p in self.TPsFPsList :
      if p['truthUID'] != None :
        n += 1
    return n

  def computeAndWriteFrocCurves(self) :
    if path.exists(self.frocFileName) :
      raise AppException('Aborting because file already exists: %s' % self.frocFileName)
    self.computeFrocCurves(excludeDuplicateTPs=True)
    #self.computeFrocCurves(excludeDuplicateTPs=False)
    self.writeFrocCurves()

  def computeFrocCurves(self, excludeDuplicateTPs=True) :
    if excludeDuplicateTPs :
      self._numObjectsToDetect = len(self.getAllTruthUIDs()) + self.numFalseNegatives
    else :
      self._numObjectsToDetect = self.getNumTPsIncludingDuplicates()
    print ('Sorting the detection list. Number of points in the list: %s' % len(self.TPsFPsList))
    self.TPsFPsList.sort(key = lambda p: float(p['likelihood']), reverse = True)
    print ('Calculating FROC curve points.')
    previousPoint = [None, None, None]
    currentPoint = [0, 0, 1]
    numTP = 0
    numFP = 0
    numExcluded = 0
    handledTruthObjects = set()
    for p in self.TPsFPsList :
      if p['truthUID'] == None :
        ### This positive is a false positive.
        numFP += 1
        self.addPointToFrocCurves(previousPoint, currentPoint, (numFP, numTP, p['likelihood']))
      else :
        ### This positive is a true positive.
        if excludeDuplicateTPs and p['truthUID'] in handledTruthObjects :
          ### This true positive is excluded because it is a duplicate.
          numExcluded += 1
        else :
          handledTruthObjects.add(p['truthUID'])
          numTP += 1
          self.addPointToFrocCurves(previousPoint, currentPoint, (numFP, numTP, p['likelihood']))
    ### Add last point to curve:
    self.addPointToFrocCurves(currentPoint, [numFP, numTP, p['likelihood']], (None, None, None))

    print 'number of FPs: %s; number of TPs: %s; number of excluded points: %s' % (numFP, numTP, numExcluded)

    #meanFPPS = numFP / self._numStudies
    ##logger.logHTML("<b>{0} false positives in {1} scans. Mean: {2} false positives per scan</b>".format(numFP, self._numStudies, meanFPPS))
    #maxSensitivity = numTP / float(self._numObjectsToDetect)
    #logger.logHTML("<b>Detected {0} out of {1} objects (in {3} scans). Maximum sensitivity: {2}</b>".format(numTP, self._numObjectsToDetect, maxSensitivity, self._numStudies))

  def addPointToFrocCurves(self, previousPoint, currentPoint, nextPoint) :
    if ((previousPoint[0] == 0)
        or ((not previousPoint[0] == currentPoint[0] == nextPoint[0]) and (not previousPoint[1] == currentPoint[1] == nextPoint[1]))) :
      #self._frocCurve.append(( currentPoint[0] / float(self._numStudies),         currentPoint[1] / float(self._numObjectsToDetect), currentPoint[2]))
      self._rfrocCurve.append((currentPoint[0] / float(self._numObjectsToDetect), currentPoint[1] / float(self._numObjectsToDetect), currentPoint[2]))
    for i in range(3) :
      previousPoint[i] = currentPoint[i]
      currentPoint[i] = nextPoint[i]

  def printSummary(self) :
    print "printSummary: under construction"

  def writeFrocCurves(self, overwrite=False) :
    if overwrite == False and os.path.exists(self.frocFileName) :
      raise AppException('Aborting because file already exists: %s' % self.frocFileName)
    #if "froc" in self.frocFileName :
    #  rfrocFileName = self.frocFileName.replace('froc', 'rfroc') # Rational FROC
    #else :
    #  rfrocFileName = self.frocFileName + '.rfroc'
    rfrocFileName = self.frocFileName
    #for curve, fileName in ((self._frocCurve, frocFileName), (self._dfrocCurve, dfrocFileName), (self._rfrocCurve, rfrocFileName)) :
    #for curve, fileName in ((self._frocCurve, self.frocFileName), (self._rfrocCurve, rfrocFileName)) :
    for curve, fileName in ((self._rfrocCurve, rfrocFileName),) :
      print ('Writing froc curve points to file %s' % fileName)
      with open(fileName, "w") as outFile :
        for point in curve :
          textLine = "%s,%s,%s\n" % (point[0], point[1], point[2])
          #logger.log("[line %s] %s" % (i, textLine))
          outFile.write(textLine)


