# -*- coding: utf-8 -*-
"""
Created on Fri Mar 17 13:25:52 2017

@author: Oscar Debats
"""
from sys import argv
from os import path
import os
from local_tools.exceptions import AppException
from local_tools.froc_analysis import FrocCurve


class Main() :
  def __init__(self) :
    pass

  def main(self) :
    if len(argv) == 1 :
      print "Type 'python %s --help' for usage." % argv[0]
      quit()
    import argparse
    parser = argparse.ArgumentParser(description="Tools for the lymph node detection pipeline ")
    parser.add_argument('COMMAND', default="lik", choices=["lik", "likelihood_analysis", "count", "froc", "test"])
    parser.add_argument('dirs', nargs='*', default=[os.getcwd()])
    parser.add_argument('--threshold', type=float, default=0.0, help='Set a minimum threshold for likelihood value.')
    parser.add_argument('--output_dir', help='Specify the output directory.')
    self.args = parser.parse_args()
    self.command = self.args.COMMAND

    for directory in self.args.dirs :
      if not path.isdir(directory) :
        raise AppException('ERROR: Input directory not found: %s' % directory)
    if self.args.output_dir :
      if not path.isdir(self.args.output_dir) :
        raise AppException('ERROR: Output directory not found: %s' % self.args.output_dir)
      elif not os.listdir(self.args.output_dir) == [] :
        raise AppException('ERROR: Output directory not empty: %s' % self.args.output_dir)

    if self.command in ["lik", "likelihood_analysis"] :
      self.likelihoodAnalysis()
    elif self.command in ["count"] :
      self.countUniqueValues()
    elif self.command in ["froc"] :
      self.colNumber = 5 # Column in which the likelihood values are to be found.
      #self.colNumber = 2 # Column in which the likelihood values are to be found.
      self.computeFrocCurve(self.args.dirs)
    else :
      raise AppException('ERROR: Unexpected command "%s".' % self.command)
    print 'Finished.'

  def countUniqueValues(self) :
    self.colNumber = 3 # Column in which to count unique values.
    self.valueList = []
    self.processDirs()
    self.uniquenessSummary()

  def computeFrocCurve(self, dirs, falseNegativesFolder, frocFileName) :
    self.froc = FrocCurve()
    self.froc.frocFileName = frocFileName
    if path.exists(self.froc.frocFileName) :
      raise AppException('Aborting because file already exists: %s' % self.froc.frocFileName)
    self.textProcessor = LineByLineTextProcessor()
    self.valueList = self.froc.TPsFPsList
    self.textProcessor.processDirs(dirs, self.addDetectionPointFromTextLine)
    self.falseNegativesList = []
    self.colNumber = 5 # Column in which the truth number of a false negative is to be found.
    self.textProcessor.processDirs([falseNegativesFolder], self.addFalseNegativeFromTextLine)
    #print self.falseNegativesList
    print 'Number of false negatives: %s' % len(self.falseNegativesList)
    self.froc.numFalseNegatives = len(self.falseNegativesList)
    self.froc.computeAndWriteFrocCurves()
    self.froc.printSummary()

  def likelihoodAnalysis(self) :
    self.colNumber = 5 # Column in which the likelihood values are to be found.
    self.valueList = []
    self.countSkippedLikelihoods = 0
    self.processDirs()
    self.likelihoodAnalysisSummary()

  def likelihoodAnalysisSummary(self) :
    print "Collected %s likelihood values from %s files." % (len(self.valueList), self.fileCount)
    print 'Skipped  %s likelihood values that are below the threshold of %s' % (self.countSkippedLikelihoods,
                                                                                self.args.threshold)
    print 'Sorting...'
    self.valueList.sort(reverse=True)
    print 'The 10 highest likelihoods found are: %s' % self.valueList[:10]
    print 'The 10 lowest likelihoods found are: %s' % self.valueList[-10:]
    maxNumLikelihoods = self.fileCount * 1000
    if len(self.valueList) > max(0, maxNumLikelihoods-1) :
      print 'The likelihood at list position %s is %s.' % (maxNumLikelihoods-1, self.valueList[maxNumLikelihoods-1])
    if len(self.valueList) > maxNumLikelihoods :
      print 'The likelihood at list position %s is %s.' % (maxNumLikelihoods, self.valueList[maxNumLikelihoods])

  def uniquenessSummary(self) :
    print "Listing %s unique string values:" % len(self.valueList)
    for i, value in enumerate(sorted(self.valueList)) :
      print "%s:%s" % (i+1, value)
    print "Found %s unique string values." % len(self.valueList)

  def getStringValueFromCsvTextLine(self, line, colNumber=None) :
    if colNumber == None : colNumber = self.colNumber
    splitLine = line.strip().split(",")
    if len (splitLine) < colNumber :
      raise AppException('ERROR: Column %s not found while interpreting text as comma-separated values.' % colNumber)
    stringValue = splitLine[colNumber-1]
    return stringValue

  def processOneLine(self, outFile, line) :
    stringValue = self.getStringValueFromCsvTextLine(line)
    if self.command in ["lik", "likelihood_analysis"] :
      self.processOneLikelihood(outFile, line, stringValue)
    elif self.command in ["count"] :
      self.addIfUnique(stringValue)
    else :
      raise AppException('ERROR: unrecognized command: %s' % self.command)

  def addDetectionPointFromTextLine(self, outFile, line) :
      stringValue = self.getStringValueFromCsvTextLine(line, self.colNumberDict['likelihood'])
      if '_' in stringValue :
        likelihood, truthID = stringValue.split('_')
      else :
        likelihood = stringValue
        truthID = self.getStringValueFromCsvTextLine(line, 0)
      if 'studyID' in self.colNumberDict.keys() :
        s = self.getStringValueFromCsvTextLine(line, self.colNumberDict['studyID'])
        studyID = s.split('_')[1] ### Study ID should be the 2nd part of the string value in the 3rd column.
      else :
        studyID = path.split(self.textProcessor.inFileName)[-1].split('_')[1]
      if truthID == '0' :
        truthUID = None # The current point is a false positive.
      else:
        truthUID = (truthID, studyID)
      self.froc.addPoint(likelihood, truthUID)

  def addFalseNegativeFromTextLine(self, outFile, line) :
    stringValue = self.getStringValueFromCsvTextLine(line)
    falseNegativeNumber = stringValue.split('_')[1]
    studyID = path.split(self.textProcessor.inFileName)[-1].split('_')[1]
    falseNegativeID = (falseNegativeNumber, studyID)
    self.falseNegativesList.append(falseNegativeID)

  def processOneLikelihood(self, outFile, line, stringValue) :
    floatValue = float(stringValue.split("_")[0])
    if floatValue >= self.args.threshold :
      self.valueList.append(floatValue)
      if outFile :
        outFile.write(line.strip() + "\n")
    else :
      self.countSkippedLikelihoods += 1

  def addIfUnique(self, stringValue) :
    stringValue = "_".join(stringValue.split("_")[0:2])
    if stringValue not in self.valueList :
      self.valueList.append(stringValue)
      #print "%s:%s" % (len(self.valueList), stringValue)

  def addTruthNumbersToClassificationResults(self, TPsFolder, resultFolder) :
    self.textProcessor = LineByLineTextProcessor()
    self.createTruthNumberDict(TPsFolder)
    #for k in sorted(self.truthNumberDict.keys()) :
    #  print k
    ### Add truth numbers to (copies of) result files:
    dictFileName = path.join(resultFolder, 'truthNumberDict.txt')
    with open(dictFileName, 'w') as outFile :
      print 'Writing truthNumberDict to file: %s' % (dictFileName)
      for k in sorted(self.truthNumberDict.keys()) :
       outFile.write('%s : %s\n' % (k, self.truthNumberDict[k]))
    self.textProcessor.output_dir = resultFolder + '-with_truthIDs'
    self.textProcessor.processDirs([resultFolder], self.addTruthNumberToResultFile)

  def createTruthNumberDict(self, TPsFolder) :
    self.truthNumberDict = {}
    self.colNumber = 5 # Column in which the likelihood and truth number are to be found.
    self.textProcessor.processDirs([TPsFolder], self.addTruthNumberToDict)

  def addTruthNumberToDict(self, outFile, line) :
    stringValue = self.getStringValueFromCsvTextLine(line)
    truthNumber = stringValue.split('_')[1]
    studyID = path.split(self.textProcessor.inFileName)[-1].split('_')[1]
    lineNumber = self.textProcessor.lineCount - 1
    self.truthNumberDict[(studyID, lineNumber)] = truthNumber

  def addTruthNumberToResultFile(self, outFile, line) :
    splittedLine = line.strip().split(",")
    likelihood = splittedLine[1] ### Likelihood should be in the 2nd column.
    classID = splittedLine[0] ### Class ID (class 0 or class 1) should be in the 1st column.
    if classID == '0' :
      truthNumber = 0
    elif classID == '1' : ### Truth numbers apply, of course, only to class 1.
      studyID = splittedLine[2].split('_')[1] ### Study ID should be the 2nd part of the string value in the 3rd column.
      patchID = splittedLine[2].split('_')[2] ### Patch ID should be the 3rd part of the string value in the 3rd column.
      patchNumber = int(patchID[-3:])
      truthNumber = self.truthNumberDict[(studyID, patchNumber)]
    else :
      raise AppException('Unexpected class ID: "%s"' % classID)
    likelihoodAndTruthNumber = '%s_%s' % (likelihood, truthNumber)
    splittedLine[1] = likelihoodAndTruthNumber
    joinedLine = ','.join(splittedLine)
    outFile.write(joinedLine.strip() + "\n")


class LineByLineTextProcessor() :
  def __init__(self) :
    self.inFileName = ""
    self.output_dir = ""
    self.fileCount = 0
    self.lineCount = None

  def processDirs(self, dirs, lineProcessorFunction) :
    self._processOneLine = lineProcessorFunction
    for directory in dirs :
      self.processFiles(directory)
    print 'Finished processing %s directories: %s' % (len(dirs), repr(dirs))

  def processFiles(self, directory) :
    fileNames = [path.join(directory, item) for item in sorted(os.listdir(directory))
                 if path.splitext(item)[1]=='.csv']
    fileNames = [item for item in fileNames if path.isfile(item)]
    print 'Reading %s files from directory "%s"' % (len(fileNames),directory)
    for i, inFileName in enumerate(fileNames) :
      #if self.fileCount == 1 : break # Limit the number of files to read (for testing)
      outFileName = None
      if self.output_dir :
        if not path.isdir(self.output_dir) :
          raise AppException('Output directory not found: %s' % self.output_dir)
        #outFileName = path.join(self.output_dir, path.splitext(path.split(inFileName)[-1])[0] + '-filtered.csv')
        outFileName = path.join(self.output_dir, path.splitext(path.split(inFileName)[-1])[0] + '-enriched.csv')
      self.fileCount += 1
      self.processOneFile(inFileName, outFileName)
    if self.fileCount == 0 :
      print 'WARNING: No files read.'

  def processOneFile(self, inFileName, outFileName) :
    #offset = len(self.valueList)
    self.inFileName = inFileName
    with open(inFileName) as inFile :
      print 'Reading file %s: %s' % (self.fileCount, path.split(inFileName)[-1])
      firstLine = inFile.readline()
      if firstLine == "" :
        print 'WARNING: Empty file: %s' % inFileName
        # Maybe create empty output file?
        return
      if outFileName :
        with open(outFileName, 'w') as outFile :
          print 'Writing file %s: %s' % (self.fileCount, outFileName)
          self.processLines(firstLine, inFile, outFile)
      else :
        self.processLines(firstLine, inFile, outFile=None)
      #print 'Added %s likelihood values.' % (len(self.valueList) - offset)

  def processLines(self, firstLine, inFile, outFile) :
    self.lineCount = 1 # First line has been read already; its contents are in firstLine.
    self._processOneLine(outFile, firstLine)
    for line in inFile :
      self.lineCount += 1
      self._processOneLine(outFile, line)
      #if self.lineCount == 10 : break # Limit the number of lines to read (for testing)


if __name__ == '__main__' :
    # execute only if run as a script

    mainObject = Main()
    try :
      mainObject.main()
    except AppException as e :
      print e
    quit()