# -*- coding: utf-8 -*-
"""
Created on Tue Jul 07 16:02:52 2015

@author: Geert
"""

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser("Training script for prostate segmentation networks")
    parser.add_argument("--views", "-v", type=int, default=0, nargs='+', required=True, help="Which views should be used")
    parser.add_argument("--save_dir", "-d", type=str, default="", required=True)
    parser.add_argument("--learning_rate", "-l", type=float, default=0.0001, help="Learning rate")
    parser.add_argument("--local_path", "-p", type=str, default="", required=False, help="When not running on a deepr machine, specify the root path for the folder with the data and code.")
    parser.add_argument("--dropout", "-o", type=float, default=0.0, help="Probability for a value in the layers to be set to zero.")
    parser.add_argument("--l2_constant", "-c", type=float, default=0.0, help="Set the amount of weight regularization during training.")
    parser.add_argument("--num_epochs", "-e", type=int, default=1000, help="Set the number of training epochs.")
    parser.add_argument("--anneal", "-a", action='store_true', help="Anneal the learning rate after 10 epochs of not improving the validation loss.")

    args = parser.parse_args()

    import os
    views = args.views
    model_save_folder = os.path.realpath(args.save_dir)
    learning_rate = args.learning_rate
    local_path = os.path.realpath(args.local_path)
    dropout_p = args.dropout
    l2_constant = args.l2_constant
    num_epochs = args.num_epochs
    if args.anneal:
        learning_rate_decay = 0.9
    else:
        learning_rate_decay = None

    patch_size = 65

    import joblib
    import theano
    import theano.tensor as T
    import lasagne.layers as L
    from lasagne.regularization import regularize_network_params, l2
    import lasagne
    import os
    import numpy as np
    rng_state = np.random.RandomState(1234)
    lasagne.random.set_rng(rng_state)

    base_path = r"/mnt/resfilsp10/prostate/ProjectData/Lymphography" #base_path = r"\\resfilsp10\imaging\prostate\ProjectData\Lymphography"
    if local_path :
        #base_path = local_path
        data_pth = os.path.join(local_path, r"iterator_structure_2")
        data_pth = "/mnt/SSD/oscar/data"
    else :
        #data_pth = os.path.join(base_path, r"exp/dl/iterator_structure_2")
        data_pth = "/mnt/SSD/oscar/data"
    print "Base path:", base_path
    #data_pth = os.path.join(base_path, "exp", "dl", "iterator_structure")

    print "Data path:", data_pth

    python_path1 = os.path.join(base_path, "exp", "dl_software", "python")
    python_path2 = os.path.join(base_path, "exp", "dl_software", "python", "training")
    print python_path1; print python_path2

    import sys
    sys.path.append(python_path1)
    sys.path.append(python_path2)
    sys.setrecursionlimit(1000000)

    from deeplearning.data_handling.batch_iterator import BatchIterator
    from deeplearning.data_handling.patch_batch_provider import MultiScalePatchBatchProvider
    from tmi_evaluation.training.functions import train

    print "Compiling network..."
    target_var = T.fmatrix('targets')
    prev_model_files = sorted([x for x in os.listdir(model_save_folder) if x.endswith(".pkl")])
    if not prev_model_files:
        input_vars = [T.tensor4(theano.config.floatX) for view in views]
        from tmi_evaluation.models.complex_multiview import cnn
        cnn_def = cnn(input_vars, dropout_p=dropout_p)

        prev_epoch_dict  = None
        ls = lasagne.layers.get_all_layers(cnn_def)
        input_vars = [x.input_var for x in ls if isinstance(x, L.InputLayer)]
    else:
        prev_epoch_dict = joblib.load(os.path.join(model_save_folder, prev_model_files[-1]))
        cnn_def = prev_epoch_dict['model']
        ls = lasagne.layers.get_all_layers(cnn_def)
        input_vars = [x.input_var for x in ls if isinstance(x, L.InputLayer)]

    prediction = T.nnet.softmax(T.clip(T.flatten(lasagne.layers.get_output(cnn_def), 2), 0.00001, 0.99999))
    train_loss = lasagne.objectives.categorical_crossentropy(prediction, target_var)
    train_loss = train_loss.mean()
    train_acc = T.mean(T.eq(T.argmax(prediction, axis=1), T.argmax(target_var, axis=1)),
                      dtype=theano.config.floatX)
    train_outputs = [train_loss, train_acc, prediction]

    if l2_constant > 0.:
        l2_loss = regularize_network_params(cnn_def, l2) * l2_constant
        train_loss = train_loss + l2_loss
        train_outputs = [train_loss, train_acc, prediction, l2_loss]

    params = lasagne.layers.get_all_params(cnn_def, trainable=True)
    eta = theano.shared(np.array(learning_rate, dtype=theano.config.floatX))
    eta_decay = np.array(0.75, dtype=theano.config.floatX)
    updates = lasagne.updates.nesterov_momentum(train_loss, params, learning_rate=learning_rate, momentum=0.9)

    test_prediction = T.nnet.softmax(T.clip(T.flatten(lasagne.layers.get_output(cnn_def, deterministic=True), 2), 0.00001, 0.99999))
    test_loss = lasagne.objectives.categorical_crossentropy(test_prediction, target_var)
    test_loss = test_loss.mean()

    test_acc = T.mean(T.eq(T.argmax(test_prediction, axis=1), T.argmax(target_var, axis=1)),
                      dtype=theano.config.floatX)

    train_fn = theano.function(input_vars + [target_var], train_outputs, updates=updates)
    val_fn   = theano.function(input_vars + [target_var], [test_loss, test_acc, test_prediction])
    fold = {}
    fold[1] = ['p4585711', 'p0627813', 'p8686232', 'p4532819', 'p3601996', 'p4120867', 'p0666938', 'p8598678', 'p3627815', 'p0597969', 'p6597671', 'p6599873', 'p0605393', 'p6611930', 'p3600518', 'p3597891', 'p9620177', 'p2644371', 'p3576002', 'p4644695', 'p5605439', 'p9626312', 'p4631336', 'p0594113', 'p2590891', 'p3297069', 'p6683359', 'p8665696', 'p9681507', 'p5674891', 'p1688847', 'p4687734', 'p4607937', 'p6569024', 'p2582008', 'p3593641', 'p0597811', 'p7581913', 'p0598717', 'p2574064', 'p2617602', 'p7532885', 'p7578035', 'p3432460', 'p6670262', 'p2672711', 'p3622889', 'p7641957']
    fold[2] = ['p7617043', 'p1689627', 'p6861672', 'p2663445', 'p2579805', 'p6390427', 'p6595596', 'p9569008', 'p8641446', 'p5689630', 'p9613343', 'p2613492', 'p3678125', 'p5594632', 'p8659059', 'p0600265', 'p5118883', 'p1568588', 'p2538395', 'p4570743', 'p5683896', 'p1581409', 'p4685590', 'p6576616', 'p7748634', 'p2671824', 'p8651383', 'p2686556', 'p9623970', 'p2567455', 'p3587731', 'p4570539', 'p6627402', 'p1623292', 'p9570404', 'p1599528', 'p0564317', 'p7217227', 'p7602130', 'p9685428', 'p9690928', 'p1605896', 'p0664763', 'p6683382', 'p3604916', 'p5621677', 'p9638095', 'p0663568']
    fold[3] = ['p8540654', 'p3630336', 'p0611824', 'p1433584', 'p2581432', 'p8617147', 'p5433626', 'p2582751', 'p8624871', 'p5581907', 'p0568800', 'p4576634', 'p3678775', 'p8617278', 'p3568697', 'p8603616', 'p1642985', 'p1628695', 'p6593173', 'p7601964', 'p7593337', 'p5673127', 'p1581914', 'p2597643', 'p8685703', 'p4566748', 'p5587730', 'p7624729', 'p3606942', 'p7626316', 'p4570982', 'p6660933', 'p2613698', 'p4593828', 'p6613340', 'p1575968', 'p4619845', 'p3532596', 'p0591495', 'p6690095', 'p3657142', 'p3663506', 'p8682574', 'p3593179', 'p6581500', 'p5112907', 'p5484703', 'p2668967']
    fold[4] = ['p0650020', 'p8645534', 'p3604318', 'p2611376', 'p8567244', 'p7660638', 'p0633740', 'p2685758', 'p4569432', 'p0652344', 'p7665697', 'p3400656', 'p4576318', 'p9576631', 'p1568780', 'p0662965', 'p7400205', 'p7485063', 'p9690930', 'p5623302', 'p5949007', 'p1687563', 'p3594567', 'p3694962', 'p2576475', 'p2634620', 'p3651204', 'p1685483', 'p5613053', 'p1652475', 'p0661483', 'p2632391', 'p3483299', 'p6675462', 'p1410777', 'p3600052', 'p4677934', 'p7677948', 'p2568599', 'p2583254', 'p3646005', 'p7582010', 'p7442211', 'p5551835', 'p5611719', 'p0593136', 'p4600501', 'p8616456']
    fold[5] = ['p2626311', 'p7563559', 'p6605770', 'p5688859', 'p1664761', 'p0561271', 'p1663448', 'p0617992', 'p7638076', 'p4458326', 'p9686348', 'p2662022', 'p8663444', 'p8567472', 'p0624857', 'p0630505', 'p1624744', 'p3623348', 'p1604392', 'p9595660', 'p6636335', 'p2645313', 'p6576006', 'p4682568', 'p0929146', 'p3638579', 'p3631373', 'p8432150', 'p4589872', 'p7477062', 'p6597374', 'p3683094', 'p6604925', 'p3612753', 'p7605833', 'p9616891', 'p6571820', 'p1648860', 'p7634850', 'p8652832', 'p6582013', 'p5636329', 'p3622854', 'p9571868', 'p0601079', 'p3581903', 'p0344302', 'p2664715']

    tra_fls= sorted([x for x in os.listdir(data_pth) if int(x) in views])
    for s in tra_fls :
      print "%s/%s - %s files." % (data_pth, s, len(os.listdir(os.path.join(data_pth, s))))
    val_fls = sorted([x for x in os.listdir(data_pth) if int(x) in views])

    print "Starting iterators..."
    print tra_fls
    print val_fls
    tra_bi = BatchIterator(data_pth, batch_size=32, nr_of_cached_samples=300, nr_classes=2, provider=MultiScalePatchBatchProvider,
                           balanced=True, random_state=rng_state, patch_size=[1, patch_size, patch_size], shuffled=True, scales=tra_fls,
                           #filters=fold[2] + fold[3] + fold[4])
                           filters=fold[3] + fold[4] + fold[5])
                           #filters=fold[4] + fold[5] + fold[1])
                           #filters=fold[5] + fold[1] + fold[2])
                           #filters=fold[1] + fold[2] + fold[3])
    val_bi = BatchIterator(data_pth, batch_size=32, nr_of_cached_samples=300, nr_classes=2, provider=MultiScalePatchBatchProvider,
                           balanced=True, random_state=rng_state, patch_size=[1, patch_size, patch_size], shuffled=True, scales=val_fls,
                           #filters=fold[5])
                           filters=fold[1])
                           #filters=fold[2])
                           #filters=fold[3])
                           #filters=fold[4])
    #print "Validation: fold5. Left for testing: fold1."
    #print "Validation: fold1. Left for testing: fold2."
    print "Validation: fold2. Left for testing: fold3."
    #print "Validation: fold3. Left for testing: fold4."
    #print "Validation: fold4. Left for testing: fold5."

    tra_bi.start()
    val_bi.start()
    print "Starting training process"
    try:
        train(cnn_def, train_fn, val_fn, num_epochs, 200, 200, [tra_bi], [val_bi],
              model_save_folder=model_save_folder,
              prev_epoch_dict=prev_epoch_dict, learning_rate=learning_rate,
              learning_rate_decay=learning_rate_decay)
    except KeyboardInterrupt:
        print "Terminating processes"
        tra_bi.terminate()
        val_bi.terminate()
        print "Processes terminated"
    tra_bi.terminate()
    val_bi.terminate()

    print "Finished. Data path:", data_pth
    print "Model save folder:", model_save_folder
