# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 18:31:01 2015

@author: Geert
"""

import argparse
parser = argparse.ArgumentParser("Training script for single scale networks")
parser.add_argument("--model_path", "-m", type=str, default="", required=True, help="Path to the model file to evaluate")
parser.add_argument("--output_path", "-o", type=str, default="", required=True, help="Folder to store output results")
parser.add_argument("--test", "-t", action='store_true', help="Classify the test data")
parser.add_argument("--local_path", "-lp", type=str, default="", required=False, help="When not running on a deepr machine, specify the root path for the folder with the data and code.")
args = parser.parse_args()

local_path = args.local_path

base_path = r"/mnt/resfilsp10/pathology/Userdata/geert/"
if local_path:
    base_path = local_path

import os
import joblib
import csv
import sys
sys.path.append(os.path.join(base_path, "code"))
sys.path.append(os.path.join(base_path, "code", "pathology", "install" , "bin"))
sys.path.append(os.path.join(base_path, "experiments", "TMIDeepLearning", "Python"))
sys.setrecursionlimit(1000000)
if not local_path:
    sys.path.insert(0,'/home/diag/geert/code/theano')
    sys.path.insert(0,'/home/diag/geert/code/lasagne')
import multiresolutionimageinterface as mir
from tmi_evaluation.classification.classify_slide import classify_slide
from tmi_evaluation.metrics.dice import calculate_dice_overlap_for_wsi

def classify_validation_slides(model, out_fldr, mask_level, scales, patch_size):
    #valid_pids = ['T13-02181', 'T09-15935', 'T09-20205', 'T12-18345', 'T13-03292', 'T09-00547', 'T13-02228', 'T09-01162', 'T11-17901', 'T12-02344']
    if args.test:
        pids = ['T12-05641', 'T12-00566', 'T12-00516', 'T12-00534', 'T12-02344', 'T12-04657', 'T12-13013', 'T12-25320', 'T12-28953', 'T13-05484', 'T13-06855', 'T12-06219']
    else:
        pids = ['T12-09731', 'T12-18345', 'T13-02181', 'T13-02228', 'T13-03292', 'T12-13653']
    msk_fls = {}
    fls = {}
    val_fls = {}
    if local_path:
        img_fld_10x = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Prostatectomy", "Nijmegen", "10x_converted")
        msk_fld_10x = os.path.join(img_fld_10x, "LabelMasks")
    else:
        img_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted"
        msk_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted/LabelMasks"
    msk_fls['10x'] = os.listdir(msk_fld_10x)                  
    fls['10x'] = [x for x in os.listdir(img_fld_10x) if x[0:-4] + ".tif" in msk_fls['10x']]
    val_fls['10x'] = [val_fl for val_fl in fls['10x'] if val_fl[:9] in pids]
    msk_fls['10x'] = [os.path.join(msk_fld_10x, f[0:-4] + ".tif") for f in val_fls['10x']]
    val_fls['10x'] = [os.path.join(img_fld_10x, f) for f in val_fls['10x']]

    if local_path:
        img_fld_20x = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Prostatectomy", "Nijmegen", "20x_converted")
        msk_fld_20x = os.path.join(img_fld_20x, "LabelMasks")
    else:
        img_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted"
        msk_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted/LabelMasks"    
    msk_fls['20x'] = os.listdir(msk_fld_20x)                  
    fls['20x'] = [x for x in os.listdir(img_fld_20x) if x[0:-4] + ".tif" in msk_fls['20x']]
    val_fls['20x'] = [val_fl for val_fl in fls['20x'] if val_fl[:9] in pids]
    msk_fls['20x'] = [os.path.join(msk_fld_20x, f[0:-4] + ".tif") for f in val_fls['20x']]
    val_fls['20x'] = [os.path.join(img_fld_20x, f) for f in val_fls['20x']]
      
    r = mir.MultiResolutionImageReader()
    dices = {}
    print "Processing scales: " + str(scales)
    for key in val_fls.keys():
        for im_fl, ms_fl in zip(val_fls[key], msk_fls[key]):
            print "Processing " + im_fl + " with mask " + ms_fl
            out_pth = os.path.join(out_fldr, os.path.split(ms_fl)[1])
            ms = r.open(ms_fl)
            if not os.path.exists(out_pth):
                im = r.open(im_fl)
                classify_slide(im, scales[key], model, out_pth, ms, out_scale=None, patch_size=patch_size)
                im.close()
            lk = r.open(out_pth)
            dice = calculate_dice_overlap_for_wsi(ms, lk, mask_level[key]+2, 2, 2, range(1,255))
            if dice > 0:
                dices[im_fl] = dice
            ms.close()
            lk.close()
    return dices

mask_level = {"0-6125x-complex-256" : {'20x' : 5, '10x' : 4}, "125x" :  {'20x' : 4, '10x' : 3}, "1-25x" : {'20x' : 4, '10x' : 3}, "2-5x" : {'20x' : 3, '10x' : 2}, "5x" : {'20x' : 2, '10x' : 1}, "10x" : {'20x' : 1, '10x' : 0}, 'multires-two-scales04-complex-256' : {'20x' : 5, '10x' : 4}, 'multires-three-scales' : {'20x' : 3, '10x' : 2}, 'twoscales' : {'20x' : 4, '10x' : 3}, 'threescales' : {'20x' : 4, '10x' : 3}}
scales = {"0-6125x-complex-256" : {'20x' : [32.], '10x' : [16.]}, "125x" : {'20x' : [16.], '10x' : [8.]}, "1-25x" : {'20x' : [16.], '10x' : [8.]}, "2-5x" : {'20x' : [8.], '10x' : [4.]}, "5x" : {'20x' : [4.], '10x' : [2.]}, "10x" : {'20x' : [2.], '10x' : [1.]}, "multires-two-scales04-complex-256" : {'20x' : [2., 32.], '10x' : [1., 16.]}, "multires-three-scales" : {'20x' : [2., 8., 32.], '10x' : [1., 4., 16.]}, "twoscales" : {'20x' : [4., 16.], '10x' : [2., 8.]},"threescales" : {'20x' : [4., 8., 16.], '10x' : [2., 4., 8.]}}
model = joblib.load(args.model_path)['model']
model_fldr, model_file_name = os.path.split(args.model_path)
model_scale = model_fldr.split("_")[-1]
model_epoch = model_file_name.split("_")[-1].split(".pkl")[0]
if isinstance(model, list):
    model = model[-1]
if not os.path.isdir(args.output_path):
    os.makedirs(args.output_path)
dices = classify_validation_slides(model, args.output_path, mask_level[model_scale], scales[model_scale], patch_size=256)
with open(os.path.join(args.output_path, "dices.csv"), 'wb') as csv_file:
    writer = csv.writer(csv_file)
    writer.writerow(["Filename"] + ["Threshold (" + str(x) + "/255)" for x in range(1, 255)])
    for key, value in dices.items():
       writer.writerow([os.path.split(key)[1]] + value)
