# -*- coding: utf-8 -*-
"""
Created on Tue Jul 07 16:02:52 2015

@author: Geert
"""

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser("Training script for prostate segmentation networks")
    parser.add_argument("--levels", "-l", type=int, default=0, nargs='+', required=True, help="Resolution levels to use")
    parser.add_argument("--patch_size", "-p", type=int, default=128, required=True, help="Patch size for training")
    parser.add_argument("--save_dir", "-d", type=str, default="", required=True)
    parser.add_argument("--debug_dir", "-u", type=str, default="", required=False)
    parser.add_argument("--simple", "-s", action='store_true', help="Use the simple CNN model")
    parser.add_argument("--static_train", "-t", action='store_true', help="Use the static training set")
    parser.add_argument("--static_valid", "-v", action='store_true', help="Use the static validation set for accuracy measurements")
    parser.add_argument("--learning_rate", "-lr", type=float, default=0.0001, help="Learning rate")
    parser.add_argument("--anneal", "-a", action='store_true', help="Anneal the learning rate after 10 epochs of not improving the validation loss.")
    parser.add_argument("--local_path", "-lp", type=str, default="", required=False, help="When not running on a deepr machine, specify the root path for the folder with the data and code.")
    parser.add_argument("--booster", "-b", action='store_true', help="Use online reinforcement learning")
    parser.add_argument("--dropout", "-o", type=float, default=0.0, help="Probability for a value in the layers to be set to zero.")
    parser.add_argument("--l2_constant", "-c", type=float, default=0.0, help="Set the amount of weight regularization during training.")
    parser.add_argument("--initialized", "-i", action='store_true', help="Use initialized single scale networks")
    
    args = parser.parse_args()
   
    model_save_folder=args.save_dir
    debug_dir = args.debug_dir
    levels=args.levels
    patch_size=args.patch_size
    l2_constant = args.l2_constant
    dropout_p = args.dropout
    local_path = args.local_path
    static_train = args.static_train
    static_valid = args.static_valid
    initialized = args.initialized
    if args.anneal:
        learning_rate_decay = 0.9
    else:
        learning_rate_decay = None
    learning_rate = args.learning_rate
    
    import joblib
    import theano
    import theano.tensor as T
    import lasagne.layers as L
    from lasagne.regularization import regularize_network_params, l2
    import lasagne
    import os
    import numpy as np
    rng_state = np.random.RandomState(1234)
    lasagne.random.set_rng(rng_state)
    
    base_path = r"/mnt/resfilsp10/pathology/Userdata/geert/"
    if local_path:
        base_path = local_path
    
    import sys
    sys.path.append(os.path.join(base_path, "code"))
    sys.path.append(os.path.join(base_path, "code", "pathology", "install" , "bin"))
    sys.path.append(os.path.join(base_path, "experiments", "TMIDeepLearning", "Python"))
    sys.setrecursionlimit(1000000)
    
    from deeplearning.data_handling.batch_iterator import BatchIterator
    from deeplearning.data_handling.wsi_batch_provider import WSIBatchProvider, MultiScaleWSIBatchProvider
    from deeplearning.data_handling.hdf5_batch_provider import HDF5BatchProvider
    from deeplearning.training_extensions.boosting import OnlineBooster
    from tmi_evaluation.training.functions import train
    
    target_var = T.fmatrix('targets')
    prev_model_files = sorted([x for x in os.listdir(model_save_folder) if x.endswith(".pkl")])
    if not prev_model_files:
        input_vars = [T.tensor4(theano.config.floatX) for level in levels]
        if args.simple:
            if len(levels) > 1:
                from tmi_evaluation.models.base_multires import cnn
                cnn_def = cnn(input_vars, dropout_p=dropout_p)
            else:
                from tmi_evaluation.models.base_single_scale import cnn
                cnn_def = cnn(input_vars, dropout_p=dropout_p)
        else:
            if len(levels) > 1:
                if initialized:
                    from tmi_evaluation.models.multires_initialized import cnn                    
                else:
                    from tmi_evaluation.models.multires import cnn
                cnn_def = cnn(input_vars, dropout_p=dropout_p, in_ptch_size=patch_size)
            else:
                from tmi_evaluation.models.complex_single_scale import cnn
                cnn_def = cnn(input_vars, dropout_p=dropout_p, in_ptch_size=patch_size)
                    
        prev_epoch_dict  = None
        ls = lasagne.layers.get_all_layers(cnn_def)
        input_vars = [x.input_var for x in ls if isinstance(x, L.InputLayer)]    
    else:
        prev_epoch_dict = joblib.load(os.path.join(model_save_folder, prev_model_files[-1]))
        cnn_def = prev_epoch_dict['model']
        ls = lasagne.layers.get_all_layers(cnn_def)
        input_vars = [x.input_var for x in ls if isinstance(x, L.InputLayer)]
    
    prediction = T.nnet.softmax(T.clip(T.flatten(lasagne.layers.get_output(cnn_def), 2), 0.00001, 0.99999))
    train_loss = lasagne.objectives.categorical_crossentropy(prediction, target_var)
    train_loss = train_loss.mean()
    train_acc = T.mean(T.eq(T.argmax(prediction, axis=1), T.argmax(target_var, axis=1)),
                      dtype=theano.config.floatX)
    train_outputs = [train_loss, train_acc, prediction]
    
    if l2_constant > 0.:
        l2_loss = regularize_network_params(cnn_def, l2) * l2_constant
        train_loss = train_loss + l2_loss
        train_outputs = [train_loss, train_acc, prediction, l2_loss]
    
    params = lasagne.layers.get_all_params(cnn_def, trainable=True)
    eta = theano.shared(np.array(learning_rate, dtype=theano.config.floatX))
    eta_decay = np.array(0.75, dtype=theano.config.floatX)
    #updates = lasagne.updates.adam(train_loss, params, learning_rate=learning_rate)
    updates = lasagne.updates.nesterov_momentum(train_loss, params, learning_rate=learning_rate, momentum=0.9)    
    
    test_prediction = T.nnet.softmax(T.clip(T.flatten(lasagne.layers.get_output(cnn_def, deterministic=True), 2), 0.00001, 0.99999))
    test_loss = lasagne.objectives.categorical_crossentropy(test_prediction, target_var)
    test_loss = test_loss.mean()
    
    test_acc = T.mean(T.eq(T.argmax(test_prediction, axis=1), T.argmax(target_var, axis=1)),
                      dtype=theano.config.floatX)
                      
    train_fn = theano.function(input_vars + [target_var], train_outputs, updates=updates)
    val_fn   = theano.function(input_vars + [target_var], [test_loss, test_acc, test_prediction])
    
    ### Commented PID lists are also sliced prostate specimens
    #train_pids = ['T12-08567', 'T10-22502', 'T13-02843', 'T09-09724', 'T13-01944', 'T12-08118', 'T12-27902', 'T11-08604', 'T12-06219', 'T13-00895', 'T11-02525', 'T12-19414', 'T12-27783', 'T09-02094', 'T13-01664', 'T12-06834', 'T09-05378', 'T12-12985', 'T12-09791', 'T12-00566', 'T10-25688', 'T11-06445', 'T11-12682', 'T12-21259', 'T13-08204', 'T09-07360', 'T13-06980', 'T12-18279', 'T10-12092', 'T10-02242', 'T12-13583', 'T13-02770', 'T12-27843', 'T12-25317', 'T11-13944', 'T12-28453', 'T12-08188', 'T09-19803', 'T09-10540', 'T09-00718']
    #valid_pids = ['T13-02181', 'T09-15935', 'T09-20205', 'T12-18345', 'T13-03292', 'T09-00547', 'T13-02228', 'T09-01162', 'T11-17901', 'T12-02344']
    #test_pids =  ['T10-11651', 'T09-06145', 'T13-06855', 'T12-04657', 'T12-13013', 'T09-02945', 'T10-01303', 'T09-18887', 'T09-16382', 'T12-25320', 'T10-23057', 'T11-09493', 'T11-22480', 'T12-28953', 'T11-05235', 'T09-13977', 'T10-21465', 'T11-24126', 'T12-13653', 'T12-00534', 'T10-16714', 'T09-14602', 'T10-21344', 'T11-11959', 'T11-00924', 'T12-05641', 'T12-09731', 'T09-00391', 'T11-04986', 'T13-05484']
    
    train_pids = ['T12-06834', 'T12-08118', 'T12-08188', 'T12-08567', 'T12-09791', 'T12-12985', 'T12-13583', 'T12-18279', 'T12-19414', 'T12-21259', 'T12-25317', 'T12-27783', 'T12-27843', 'T12-27902', 'T12-28453', 'T13-00895', 'T13-01664', 'T13-01944', 'T13-02770', 'T13-02843', 'T13-06980', 'T13-08204']
    valid_pids = ['T12-09731', 'T12-18345', 'T13-02181', 'T13-02228', 'T13-03292', 'T12-13653']
    test_pids = ['T12-05641', 'T12-00566', 'T12-00516', 'T12-00534', 'T12-02344', 'T12-04657', 'T12-13013', 'T12-25320', 'T12-28953', 'T13-05484', 'T13-06855', 'T12-06219']
    biopsy_train_pids = ['T12-12495', 'T12-22503', 'T12-23155', 'T12-09338', 'T13-02161', 'T12-09237', 'T12-156922', 'T12-27932', 'T12-24277', 'T12-14315', 'T12-25477', 'T12-16521', 'T12-15344', 'T12-26038', 'T12-16072', 'T12-22580', 'T12-04670', 'T12-27262', 'T12-00735', 'T12-07810', 'T12-28032', 'T12-13648', 'T12-10941', 'T12-21294', 'T12-23658', 'T12-14395', 'T12-04116', 'T12-17135', 'T12-27846', 'T13-01004', 'T12-21287', 'T12-26116', 'T12-07161', 'T12-12263', 'T12-22810', 'T12-26696', 'T12-29361', 'T12-24894', 'T12-24821', 'T12-21286', 'T12-12243', 'T12-13306', 'T13-01059', 'T13-02802', 'T12-14913', 'T12-15307', 'T12-05287', 'T12-19593', 'T12-27877', 'T13-04435', 'T12-25457', 'T12-13711', 'T12-11436', 'T12-16350', 'T12-09294', 'T12-15852', 'T12-21980', 'T12-06985', 'T12-15325', 'T12-13068', 'T12-14237', 'T12-06337', 'T12-14869', 'T12-14928', 'T12-28548', 'T12-04669', 'T12-13125', 'T12-10654', 'T12-16805', 'T12-10355', 'T12-14806', 'T12-12247', 'T12-17932', 'T12-19576', 'T12-08639', 'T12-04157', 'T12-04649', 'T12-02647', 'T12-16772', 'T12-04410', 'T12-07556', 'T12-06335', 'T12-19550', 'T12-04695', 'T12-22569', 'T13-00342', 'T12-05174', 'T12-29338', 'T12-29362', 'T12-10638', 'T12-20129', 'T12-13721', 'T12-16297', 'T12-11349', 'T13-03878', 'T12-12458', 'T12-07795', 'T12-08654', 'T12-27856', 'T12-10329']
    biopsy_valid_pids = ['T12-04516', 'T12-12578', 'T12-19540', 'T12-29093', 'T12-04409', 'T13-00360', 'T12-04702', 'T12-15383', 'T12-15548', 'T12-10390', 'T12-05198', 'T12-04484', 'T12-26657', 'T12-08620', 'T12-09255', 'T12-27933', 'T13-03866', 'T13-05194', 'T12-19696', 'T12-22516']
    biopsy_test_pids = ['T12-14523', 'T12-0933', 'T12-26039', 'T12-03262', 'T12-20261', 'T12-05828', 'T12-18902', 'T13-02767', 'T12-27276', 'T12-05745', 'T12-24776', 'T13-00370', 'T12-27909', 'T12-21578', 'T12-16765', 'T12-03339', 'T12-13111', 'T12-17259', 'T12-24818', 'T12-11399', 'T12-06931', 'T12-20686', 'T13-02803', 'T12-07762', 'T12-23081', 'T12-24935', 'T12-07534', 'T12-07761', 'T12-26672', 'T12-27876', 'T12-03626', 'T13-02239', 'T12-04650', 'T12-21323', 'T12-15547', 'T12-14539', 'T12-22522', 'T12-24284', 'T12-14269', 'T12-13627', 'T12-06942', 'T12-21999', 'T12-26062', 'T12-16073', 'T12-06336', 'T12-21934', 'T12-29337', 'T12-28462', 'T12-26677', 'T12-06978', 'T13-02247', 'T12-23133', 'T12-27855']
    
    if local_path:
        img_fld_10x = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Prostatectomy", "Nijmegen", "10x_converted")
        msk_fld_10x = os.path.join(img_fld_10x, "LabelMasks")
    else:
        img_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted"
        msk_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted/LabelMasks"
    msk_fls10x = os.listdir(msk_fld_10x)
    fls10x = [x for x in os.listdir(img_fld_10x) if x[0:-4] + ".tif" in msk_fls10x]
    
    tra_fls_10x = [os.path.join(img_fld_10x, f) for f in [tra_fl for tra_fl in fls10x if tra_fl[:9] in train_pids]]
    val_fls_10x = [os.path.join(img_fld_10x, f) for f in [val_fl for val_fl in fls10x if val_fl[:9] in valid_pids]]
    tes_fls_10x = [os.path.join(img_fld_10x, f) for f in [tes_fl for tes_fl in fls10x if tes_fl[:9] in test_pids]]
    
    if local_path:
        img_fld_20x = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Prostatectomy", "Nijmegen", "20x_converted")
        msk_fld_20x = os.path.join(img_fld_20x, "LabelMasks")
    else:
        img_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted"
        msk_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted/LabelMasks"
    msk_fls20x = os.listdir(msk_fld_20x)                  
    fls20x = [x for x in os.listdir(img_fld_20x) if x[0:-4] + ".tif" in msk_fls20x]
    
    tra_fls_20x = [os.path.join(img_fld_20x, f) for f in [tra_fl for tra_fl in fls20x if tra_fl[:9] in train_pids]]
    val_fls_20x = [os.path.join(img_fld_20x, f) for f in [val_fl for val_fl in fls20x if val_fl[:9] in valid_pids]]
    tes_fls_20x = [os.path.join(img_fld_20x, f) for f in [tes_fl for tes_fl in fls20x if tes_fl[:9] in test_pids]]
    
    if local_path:
        biopsy_pth = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Biopsy", "tif")
        biopsy_mask_pth = os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "Biopsy", "label_masks")
    else:
        biopsy_pth = r"/mnt/SSD/geert_experiment/biopsy_data"
        biopsy_mask_pth = r"/mnt/SSD/geert_experiment/biopsy_data/label_masks"
    biopsy_fls = [x for x in os.listdir(biopsy_pth) if x.endswith(".tif")]

    tra_fls_biop = [os.path.join(biopsy_pth, f) for f in [tra_fl for tra_fl in biopsy_fls if "-".join(tra_fl.split("-")[:2]) in biopsy_train_pids]]
    val_fls_biop = [os.path.join(biopsy_pth, f) for f in [val_fl for val_fl in biopsy_fls if "-".join(val_fl.split("-")[:2]) in biopsy_valid_pids]]
    tes_fls_biop = [os.path.join(biopsy_pth, f) for f in [tes_fl for tes_fl in biopsy_fls if "-".join(tes_fl.split("-")[:2]) in biopsy_test_pids]]
    

    if local_path:
        tra_10x_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "tra_10x.hf5")]
        tra_20x_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "tra_20x.hf5")]
        tra_biop_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "tra_biop.hf5")]
        val_10x_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "val_10x.hf5")]
        val_20x_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "val_20x.hf5")]
        val_bio_hdf5_pth = [os.path.join(local_path, "experiments", "TMIDeepLearning", "Data", "HDF5", "val_biop.hf5")]
    else:
        tra_10x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/tra_10x.hf5"]
        tra_20x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/tra_20x.hf5"]
        tra_biop_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/tra_biop.hf5"]
        val_10x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_10x.hf5"]
        val_20x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_20x.hf5"]
        val_bio_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_biop.hf5"]
                       
    if static_train:
        tra_bi_10x = BatchIterator(tra_10x_hdf5_pth, batch_size=12, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
        tra_bi_20x = BatchIterator(tra_20x_hdf5_pth, batch_size=12, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
        tra_bi_biop = BatchIterator(tra_biop_hdf5_pth, batch_size=12, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
    else:
        if len(levels) > 1:
            tra_bi_10x = BatchIterator(tra_fls_10x, batch_size=6, nr_of_cached_samples=300, data_level=levels[0], nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3], multi_scale_levels=levels)
            tra_bi_20x = BatchIterator(tra_fls_20x, batch_size=6, nr_of_cached_samples=300, data_level=levels[0]+1, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+1 for x in levels])
            tra_bi_biop= BatchIterator(tra_fls_biop,batch_size=6, nr_of_cached_samples=300, data_level=levels[0]+2, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+2 for x in levels])
        else:
            tra_bi_10x = BatchIterator(tra_fls_10x, batch_size=6, nr_of_cached_samples=300, data_level=levels[0], nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3])
            tra_bi_20x = BatchIterator(tra_fls_20x, batch_size=6, nr_of_cached_samples=300, data_level=levels[0]+1, nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3])
            tra_bi_biop= BatchIterator(tra_fls_biop,batch_size=6, nr_of_cached_samples=300, data_level=levels[0]+2, nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3])

    if static_valid:
        val_bi_10x = BatchIterator(val_10x_hdf5_pth, batch_size=6, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
        val_bi_20x = BatchIterator(val_20x_hdf5_pth, batch_size=6, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
        val_bi_biop = BatchIterator(val_bio_hdf5_pth, batch_size=6, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, random_state=rng_state, levels=levels)
    else:
        if len(levels) > 1:
            val_bi_10x = BatchIterator(val_fls_10x, batch_size=6, nr_of_cached_samples=500, data_level=levels[0], nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3], multi_scale_levels=levels)
            val_bi_20x = BatchIterator(val_fls_20x, batch_size=6, nr_of_cached_samples=500, data_level=levels[0]+1, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+1 for x in levels])
            val_bi_biop= BatchIterator(val_fls_biop,batch_size=6, nr_of_cached_samples=500, data_level=levels[0]+2, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, random_state=rng_state, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+2 for x in levels])
        else:
            val_bi_20x = BatchIterator(val_fls_20x, batch_size=6, nr_of_cached_samples=500, data_level=levels[0]+1, nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3])
            val_bi_10x = BatchIterator(val_fls_10x, batch_size=6, nr_of_cached_samples=500, data_level=levels[0], nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3])
            val_bi_biop= BatchIterator(val_fls_biop,batch_size=6, nr_of_cached_samples=500, data_level=levels[0]+2, nr_classes=2, provider=WSIBatchProvider, balanced=True, random_state=rng_state, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3])            
        
    if args.booster:
        bstr = OnlineBooster(alpha=1.0, required_error=0.7, only_negatives=True)
    else:
        bstr = None
    
    tra_bi_20x.start()
    tra_bi_10x.start()
    tra_bi_biop.start()
    val_bi_20x.start()
    val_bi_10x.start()
    val_bi_biop.start()
    
    print "Starting training process"
    try:
        train(cnn_def, train_fn, val_fn, 1000, 500, 2084, [tra_bi_10x, tra_bi_20x, tra_bi_biop], [val_bi_10x, val_bi_20x, val_bi_biop], booster=bstr, model_save_folder=model_save_folder, prev_epoch_dict=prev_epoch_dict, debug_folder=debug_dir, learning_rate=learning_rate, learning_rate_decay=learning_rate_decay)
    except KeyboardInterrupt:
        print "Terminating processes"
        tra_bi_20x.terminate()
        tra_bi_10x.terminate()
        val_bi_20x.terminate()
        val_bi_10x.terminate()    
        val_bi_biop.terminate()
        tra_bi_biop.terminate()
        print "Processes terminated"
    tra_bi_20x.terminate()
    tra_bi_10x.terminate()
    val_bi_20x.terminate()
    val_bi_10x.terminate()    
    val_bi_biop.terminate()
    tra_bi_biop.terminate()