# -*- coding: utf-8 -*-
"""
Created on Tue Jul 07 16:02:52 2015

@author: Geert
"""
if __name__ == '__main__':

    import argparse
    parser = argparse.ArgumentParser("Validation script for prostate segmentation networks")
    parser.add_argument("--model_folder", "-m", type=str, required=True, help="Path to the model folder")
    parser.add_argument("--levels", "-l", type=int, default=0, nargs='+', required=True, help="Resolution levels to use")
    parser.add_argument("--epoch", "-e", type=int, default=-1, required=False, help="Epoch of the model to validate")
    parser.add_argument("--patch_size", "-p", type=int, default=128, required=True, help="Patch size for training")
    parser.add_argument("--debug_dir", "-u", type=str, default="", required=False)
    parser.add_argument("--tta", "-t", action='store_true', help="Use test time augmentation")
    parser.add_argument("--static", "-s", action='store_true', help="Use the static validation set")
    parser.add_argument("--nobiopsy", "-b", action='store_true', help="Exclude biopsy data")
    args = parser.parse_args()
    
    model_folder=args.model_folder
    epoch=args.epoch
    debug_dir = args.debug_dir
    levels=args.levels
    patch_size=args.patch_size
    tta = args.tta
    nobiopsy = args.nobiopsy
    static = args.static
#==============================================================================
#     static = True
#     levels = [0, 2, 4]
#     patch_size = 128
#==============================================================================
    
    import joblib
    import theano
    import theano.tensor as T
    import lasagne.layers as L
    import lasagne
    import os
    
    base_path = r"/mnt/resfilsp10/pathology/Userdata/geert/"    
    import sys
    sys.path.append(os.path.join(base_path, "code"))
    sys.path.append(os.path.join(base_path, "code", "pathology", "install" , "bin"))
    sys.path.append(os.path.join(base_path, "experiments", "TMIDeepLearning", "Python"))
    sys.setrecursionlimit(1000000)
    
    from deeplearning.data_handling.batch_iterator import BatchIterator
    from deeplearning.data_handling.wsi_batch_provider import WSIBatchProvider, MultiScaleWSIBatchProvider
    from deeplearning.data_handling.hdf5_batch_provider import HDF5BatchProvider
    from tmi_evaluation.training.functions import validate
    
    target_var = T.fmatrix('targets')
    prev_model_files = sorted([x for x in os.listdir(model_folder) if x.endswith(".pkl")])
    if not prev_model_files:
        raise IndexError("No model files present")
    else:
        if epoch < 0:
            prev_epoch_dict = joblib.load(os.path.join(model_folder, prev_model_files[0]))
        else:
            prev_epoch_dict = joblib.load(os.path.join(model_folder, prev_model_files[epoch]))
        cnn_def = prev_epoch_dict['model']
        ls = lasagne.layers.get_all_layers(cnn_def)
        input_vars = [x.input_var for x in ls if isinstance(x, L.InputLayer)]
    
    test_prediction = T.nnet.softmax(T.clip(T.flatten(lasagne.layers.get_output(cnn_def, deterministic=True), 2), 0.00001, 0.99999))
    test_loss = lasagne.objectives.categorical_crossentropy(test_prediction, target_var)
    test_loss = test_loss.mean()
    
    test_acc = T.mean(T.eq(T.argmax(test_prediction, axis=1), T.argmax(target_var, axis=1)),
                      dtype=theano.config.floatX)
    val_fn   = theano.function(input_vars + [target_var], [test_loss, test_acc, test_prediction])
    
    ### Commented PID lists are also sliced prostate specimens
    #train_pids = ['T12-08567', 'T10-22502', 'T13-02843', 'T09-09724', 'T13-01944', 'T12-08118', 'T12-27902', 'T11-08604', 'T12-06219', 'T13-00895', 'T11-02525', 'T12-19414', 'T12-27783', 'T09-02094', 'T13-01664', 'T12-06834', 'T09-05378', 'T12-12985', 'T12-09791', 'T12-00566', 'T10-25688', 'T11-06445', 'T11-12682', 'T12-21259', 'T13-08204', 'T09-07360', 'T13-06980', 'T12-18279', 'T10-12092', 'T10-02242', 'T12-13583', 'T13-02770', 'T12-27843', 'T12-25317', 'T11-13944', 'T12-28453', 'T12-08188', 'T09-19803', 'T09-10540', 'T09-00718']
    #valid_pids = ['T13-02181', 'T09-15935', 'T09-20205', 'T12-18345', 'T13-03292', 'T09-00547', 'T13-02228', 'T09-01162', 'T11-17901', 'T12-02344']
    #test_pids =  ['T10-11651', 'T09-06145', 'T13-06855', 'T12-04657', 'T12-13013', 'T09-02945', 'T10-01303', 'T09-18887', 'T09-16382', 'T12-25320', 'T10-23057', 'T11-09493', 'T11-22480', 'T12-28953', 'T11-05235', 'T09-13977', 'T10-21465', 'T11-24126', 'T12-13653', 'T12-00534', 'T10-16714', 'T09-14602', 'T10-21344', 'T11-11959', 'T11-00924', 'T12-05641', 'T12-09731', 'T09-00391', 'T11-04986', 'T13-05484']
    
    if not static:
        valid_pids = ['T12-09731', 'T12-18345', 'T13-02181', 'T13-02228', 'T13-03292', 'T12-13653']
        biopsy_valid_pids = ['T12-04516', 'T12-12578', 'T12-19540', 'T12-29093', 'T12-04409', 'T13-00360', 'T12-04702', 'T12-15383', 'T12-15548', 'T12-10390', 'T12-05198', 'T12-04484', 'T12-26657', 'T12-08620', 'T12-09255', 'T12-27933', 'T13-03866', 'T13-05194', 'T12-19696', 'T12-22516']
        
        img_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted"
        msk_fld_10x = r"/mnt/SSD/geert_experiment/10x_converted/LabelMasks"
        msk_fls10x = os.listdir(msk_fld_10x)
        fls10x = [x for x in os.listdir(img_fld_10x) if x[0:-4] + ".tif" in msk_fls10x]        
        val_fls_10x = [os.path.join(img_fld_10x, f) for f in [val_fl for val_fl in fls10x if val_fl[:9] in valid_pids]]
        
        img_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted"
        msk_fld_20x = r"/mnt/SSD/geert_experiment/20x_converted/LabelMasks"
        msk_fls20x = os.listdir(msk_fld_20x)                  
        fls20x = [x for x in os.listdir(img_fld_20x) if x[0:-4] + ".tif" in msk_fls20x]        
        val_fls_20x = [os.path.join(img_fld_20x, f) for f in [val_fl for val_fl in fls20x if val_fl[:9] in valid_pids]]
        
        biopsy_pth = r"/mnt/diskA/geert/biopsy_data/tif"
        biopsy_mask_pth = r"/mnt/diskA/geert/biopsy_data/label_masks"
        biopsy_fls = [x for x in os.listdir(biopsy_pth) if x.endswith(".tif")]
        val_fls_biop = [os.path.join(biopsy_pth, f) for f in [val_fl for val_fl in biopsy_fls if "-".join(val_fl.split("-")[:2]) in biopsy_valid_pids]]
        
        if len(levels) > 1:
            val_bi_10x = BatchIterator(val_fls_10x, batch_size=16, nr_of_cached_samples=500, data_level=levels[0], nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3], multi_scale_levels=levels)
            val_bi_20x = BatchIterator(val_fls_20x, batch_size=16, nr_of_cached_samples=500, data_level=levels[0]+1, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+1 for x in levels])
            val_bi_biop= BatchIterator(val_fls_biop,batch_size=16, nr_of_cached_samples=500, data_level=levels[0]+2, nr_classes=2, provider=MultiScaleWSIBatchProvider, balanced=True, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3], multi_scale_levels=[x+2 for x in levels])
        else:
            val_bi_10x = BatchIterator(val_fls_10x, batch_size=16, nr_of_cached_samples=500, data_level=levels[0], nr_classes=2, provider=WSIBatchProvider, balanced=True, msk_src=msk_fld_10x, masks_level=6, patch_size=[patch_size,patch_size,3])
            val_bi_20x = BatchIterator(val_fls_20x, batch_size=16, nr_of_cached_samples=500, data_level=levels[0]+1, nr_classes=2, provider=WSIBatchProvider, balanced=True, msk_src=msk_fld_20x, masks_level=7, patch_size=[patch_size,patch_size,3])
            val_bi_biop= BatchIterator(val_fls_biop,batch_size=16, nr_of_cached_samples=500, data_level=levels[0]+2, nr_classes=2, provider=WSIBatchProvider, balanced=True, msk_src=biopsy_mask_pth, masks_level=4, patch_size=[patch_size,patch_size,3])        
        validate_batches = 500
    else:
        if not "128" in model_folder:
            val_10x_hdf5_pth = [r"/mnt/diskA/geert/val_10x.hf5"]
            val_20x_hdf5_pth = [r"/mnt/diskA/geert/val_20x.hf5"]
            val_bio_hdf5_pth = [r"/mnt/diskA/geert/val_biop.hf5"]
        else:
            val_10x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_10x.hf5"]
            val_20x_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_20x.hf5"]
            val_bio_hdf5_pth = [r"/mnt/SSD/geert_experiment/hdf5/val_biop.hf5"]
        
        val_bi_10x = BatchIterator(val_10x_hdf5_pth, batch_size=16, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, levels=levels)
        val_bi_20x = BatchIterator(val_20x_hdf5_pth, batch_size=16, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, levels=levels)
        val_bi_biop = BatchIterator(val_bio_hdf5_pth, batch_size=16, nr_of_cached_samples=500, nr_classes=2, provider=HDF5BatchProvider, balanced=True, levels=levels)
        validate_batches = 2084
    val_bi_20x.start()
    val_bi_10x.start()
    if not nobiopsy:
        val_bi_biop.start()
    
    if nobiopsy:
        val_its = [val_bi_20x, val_bi_10x]
    else:
        val_its = [val_bi_20x, val_bi_10x, val_bi_biop]
    print "Starting validation process"
    try:
        if epoch < 0:
            val_accs = []
            for cur_epoch in xrange(len(prev_model_files)):
                print "Validating epoch " + str(cur_epoch)
                cur_epoch_dict = joblib.load(os.path.join(model_folder, prev_model_files[cur_epoch]))
                cur_cnn_def = cur_epoch_dict['model']
                params = L.get_all_param_values(cur_cnn_def)
                L.set_all_param_values(cnn_def, params)
                _, cur_val_acc = validate(cnn_def, val_fn, cur_epoch, validate_batches, val_its, debug_folder=debug_dir)
                val_accs.append(cur_val_acc)
            print "Accuracies: "
            print val_accs
        else:
            validate(cnn_def, val_fn, epoch, validate_batches, val_its, debug_folder=debug_dir, tta=tta)
    except KeyboardInterrupt:
        print "Terminating processes"
        val_bi_20x.terminate()
        val_bi_10x.terminate()    
        val_bi_biop.terminate()
        print "Processes terminated"
    val_bi_20x.terminate()
    val_bi_10x.terminate()    
    val_bi_biop.terminate()
