import lasagne
from lasagne.layers.normalization import batch_norm
from lasagne.layers import dropout

def cnn(input_vars, dropout_p=0.0):
    scale_layers = []
    nr_filters = 24
    for iv in input_vars:
        network = lasagne.layers.InputLayer(shape=(None, 1, None, None), input_var=iv)
        network = lasagne.layers.Conv2DLayer(
                  network, num_filters=nr_filters, filter_size=(5, 5),
                  nonlinearity=lasagne.nonlinearities.rectify,
                  W=lasagne.init.GlorotUniform())
        network = lasagne.layers.MaxPool2DLayer(network, pool_size=(2, 2))
        if dropout_p > 0.:
            network = dropout(network, p=dropout_p)                          
        network = lasagne.layers.Conv2DLayer(
                  network, num_filters=2*nr_filters, filter_size=(5, 5),
                  nonlinearity=lasagne.nonlinearities.rectify,
                  W=lasagne.init.GlorotUniform())
        network = lasagne.layers.MaxPool2DLayer(network, pool_size=(2, 2))
        scale_layers.append(network)
    network = lasagne.layers.ConcatLayer(scale_layers)
    if dropout_p > 0.:
        network = dropout(network, p=dropout_p)    
    network = lasagne.layers.Conv2DLayer(
              network,
              num_filters=512, filter_size=(13, 13),
              nonlinearity=lasagne.nonlinearities.rectify,
              W=lasagne.init.GlorotUniform())
    if dropout_p > 0.:
        network = dropout(network, p=dropout_p)              
    network = batch_norm(lasagne.layers.Conv2DLayer(
              network,
              num_filters=2, filter_size=(1, 1),
              nonlinearity=lasagne.nonlinearities.identity,
              W=lasagne.init.GlorotUniform()))
    return network